/*		 ITI-FG  Kernel module free public release
 *
 *		     Copyright (C) 1996-2000 GOM mbH.
 *	         Rebenring 33, 38106 Braunschweig, GERMANY
 *		            All rights reserved
 *
 * $Id: itifgIOP.c,v 7.0 2000/10/25 09:07:57 cvs Exp $
 *
 */

/*
 * Defines:
 */

/*
 * Includes:
 */

#include "itifgOS.h"	/* os specific stuff */
#include "itifgExt.h"	/* offical driver interface */
#include "itifgReg.h"	/* register definitions ICR Regs */
#include "icpciReg.h"	/* register definitions IC-PCI */
#include "amvsReg.h"	/* register definitions AM-VS */
#include "amdigReg.h"	/* register definitions AM-DIG */
#include "ampcvReg.h"	/* register definitions PCVision */
#include "pcdigReg.h"	/* register definitions PC-DIG */
#include "itifgIntD.h"	/* driver internals */
#include "itifgIntF.h"	/* internal functions */

/*
 * Globals
 */

/*
 * iti_iop_interrupt
 */

void
iti_iop_interrupt (itifg_t *iti)
{
  iti->acc.events_captured++;

  if (iti->buffer != NULL && !HAS_REACHED(iop))
    {
      iti->acc.bytes_valid++;
      iti->buffer[INCR_ACNT(iop)] =
	iti->get_inport (iti);
      ITI_PRINT6A("io  captured %d: %d", ACNT(iop),
		  iti->buffer[ACNT(iop)]);
      if (HAS_REACHED(iop))
	{
	  if (!iti_block_wakeup (iti, ITI_IOPORT))
	    ITI_PRINT4("io    status: block wakeup\n");
	  else
	    iti_printn ("Nobody is waiting.\n");
	}
    }
  else
    {
      ITI_PRINT6("io  lost\n");
      iti->acc.bytes_lost++;
    }
}

/*
 * iti_iop_timeout
 */

void
iti_iop_timeout (itifg_t *iti)
{
  ITI_PRINT1("> timer (io)\n");
  
  iti->iop_timeout = TRUE;

  if (!iti_block_wakeup (iti, ITI_IOPORT))
    ITI_PRINT4("io    status: block wakeup\n");
}

/*
 * iti_iop_open
 */

int
iti_iop_open (itifg_t *iti,
	      iti_file_t *file)
{
  ITI_PRINT1("> iti_iop_open\n");

  if (IS_ICPCI && IS_REV_A)
    return -ENXIO;

  if (((F_FLAGS(file) & O_ACCMODE) == O_WRONLY) ||
      ((F_FLAGS(file) & O_ACCMODE) == O_RDWR))
    {
      if (iti->write_lock)
	return -EBUSY;
      /* write: make sure, that the output port strobe pin is high */
      CHECK_FOR_EIO (iti->set_outstrb, HIGH);
      iti->write_lock = TRUE;
      /* no interrupts used here */
    }

  if (((F_FLAGS(file) & O_ACCMODE) == O_RDONLY) ||
      ((F_FLAGS(file) & O_ACCMODE) == O_RDWR))
    {
      if (iti->iop_read_lock)
	return -EBUSY;
      /* read: distinguish between raw or latched mode (with interrupts) */
      if (F_FLAGS(file) & O_NONBLOCK)
	{
	  /* set to raw mode */
	  CHECK_FOR_EIO (iti->set_inregenb, CR_DISABLE);
	}
      else
	{
	  /* set to latched mode */
	  CHECK_FOR_EIO (iti->set_inregenb, CR_ENABLE);
	  /* clear the input port buffer */
	  CHECK_FOR_EIO (iti->set_inregclr, CR_ENABLE);
	  CHECK_FOR_EIO (iti->set_inregclr, CR_DISABLE);
	  /* we expect a h-l edge at the input port strobe pin */
	  CHECK_FOR_EIO (iti->set_instrbpol, CR_INSTRB_ACT_LOW);
	  /* the interrupt pin should be connect with the strobe pin */
	  CHECK_FOR_EIO (iti->set_inpintpol, CR_INPINT_ACT_LOW);
	}
      /* if the main interrupts not already enabled */
      if (!iti->cam_read_lock)
	iti_enable_irq (iti, ITI_EIRQ_MAIN_OPEN);
      /* special irq's and the lock */
      iti_enable_irq (iti, ITI_EIRQ_PORT_OPEN);
      iti->iop_read_lock = TRUE;
      /* reset statistics */
      iti->acc.events_captured = 0UL;
      iti->acc.bytes_valid = 0UL;
      iti->acc.bytes_lost = 0UL;
    }

  ITI_PRINT1("< iti_iop_open\n");
  return OK;
}

/*
 * iti_iop_close
 */

int
iti_iop_close (itifg_t *iti,
	       iti_file_t *file)
{
  ITI_PRINT1("> iti_iop_close\n");

  /* disable special irq's */
  iti_disable_irq (iti, ITI_DIRQ_PORT_CLOSE);
  /* is this the last close */
  if (!iti->cam_read_lock)
    /* disable used interrupts */
    iti_disable_irq (iti, ITI_DIRQ_MAIN_CLOSE); 
  /* ... and timers */
  iti_del_timer (iti, ITI_IOPORT);

  if (F_FLAGS(file) & O_WRONLY || F_FLAGS(file) & O_RDWR)
    iti->write_lock = FALSE;
  if (F_FLAGS(file) & O_RDONLY || F_FLAGS(file) & O_RDWR)
    iti->iop_read_lock = FALSE;

  ITI_PRINT1("< iti_iop_close\n");  
  return OK;
}

/*
 * iti_iop_read
 */

iti_ssize_t
iti_iop_read (itifg_t *iti,
	      iti_file_t *file,
	      char *buffer,
	      size_t count)
{
  int error;

  ITI_PRINT1A("> iti_iop_read: at 0x%p, " SIZE_F " byte(s).\n",
	      buffer, count);
  
  iti->buffer = iti_malloc (count, ITI_PAGED);
  if (iti->buffer == NULL)
    return -EAGAIN;
  ITI_PRINT3A("     iti->iop->buffer: 0x%p\n", iti->buffer);

  iti->iop_desired = count;
  ZERO_ACNT(iop);

  /* distinguish between raw or latched mode (with interrupts) */
  if (F_FLAGS(file) & O_NONBLOCK)
    while (!HAS_REACHED(iop))
      iti->buffer[INCR_ACNT(iop)] =
	iti->get_inport (iti);
  else
    {
      iti->iop_timeout = FALSE;
      ITI_PRINT4("io    status: sleep (io)\n");
      if ((error = iti_sleep (iti, ITI_WAIT_IOP, ITI_IOPORT)))
	return error; /* ERESTARTSYS */
      if (iti->iop_timeout)
	return -ENODATA;
    }
  
  if ((error = iti_copyout (buffer, iti->buffer, ACNT(iop))))
    return error;
  ITI_PRINT3A("     iti->iop->buffer: 0x%p\n", iti->buffer);
  iti_free (iti->buffer, ITI_PAGED);
  iti->buffer = NULL;
  iti->iop_desired = 0;

  ITI_PRINT1("< iti_iop_read\n");    
  return ACNT(iop);
}

/*
 * iti_iop_write
 */
	
int
iti_iop_write (itifg_t *iti,
	       iti_file_t *file,
	       const char *buffer,
	       size_t count)
{
  int error;
  caddr_t iop_buffer;
  size_t iop_desired;
  iti_ssize_t iop_actual;

  ITI_PRINT1A("> iti_iop_write: at 0x%p, " SIZE_F " byte(s).\n",
	      buffer, count);
  
  iop_buffer = iti_malloc (count, ITI_PAGED);
  if (iop_buffer == NULL)
    return -EAGAIN;
  ITI_PRINT3A("     iop_buffer: 0x%p\n", iti->buffer);

  iop_desired = count;
  iop_actual = 0; /* COUNT_IS_ZERO */

  if ((error = iti_copyin (iop_buffer, buffer, count)))
    return error;

  /* write the all data as fast as possible to the outport */
  /* later we should implement a adjustable data rate here */
  while (iop_actual < iop_desired) /* !HAS_COUNT_REACHED */
    {
      CHECK_FOR_EIO (iti->set_outport,
		     iop_buffer[iop_actual++]); /* INCREMENT_COUNT */
      /* initiate a h-l edge at strobe pin to mark the data valid */
      CHECK_FOR_EIO (iti->set_outstrb, LOW);
      CHECK_FOR_EIO (iti->set_outstrb, HIGH);
    }
  iti_free (iop_buffer, ITI_PAGED);
  ITI_PRINT3A("     iop_buffer: 0x%p\n", iti->buffer);

  ITI_PRINT1("< iti_iop_write\n");    
  return iop_actual; /* COUNT */
}

